from django.urls import path
from rest_framework import routers

from .views import LayerViewset, LayerViews

app_name = "terralayer"

router = routers.SimpleRouter()

router.register(r"", LayerViewset, base_name="layer")

urlpatterns = [
    path(r"view/", LayerViews.as_view(), name="layerview"),
    path(r"view/<str:slug>/", LayerViews.as_view(), name="layerview"),
]

urlpatterns += router.urls



from rest_framework import permissions


class LayerPermission(permissions.BasePermission):
    def has_permission(self, request, view):
        return request.user.has_perm("terra_layer.can_manage_layers")



from rest_framework.serializers import (
    ModelSerializer,
    PrimaryKeyRelatedField,
    ValidationError,
)

from django.db import transaction

from .models import Layer, LayerGroup, FilterField, CustomStyle


class FilterFieldSerializer(ModelSerializer):
    id = PrimaryKeyRelatedField(source="field", read_only=True)

    class Meta:
        model = FilterField
        exclude = ("layer",)


class CustomStyleSerializer(ModelSerializer):
    class Meta:
        model = CustomStyle
        exclude = ("layer",)


class LayerSerializer(ModelSerializer):
    fields = FilterFieldSerializer(many=True, read_only=True, source="fields_filters")
    custom_styles = CustomStyleSerializer(many=True, read_only=True)

    @transaction.atomic
    def create(self, validated_data):
        instance = super().create(validated_data)

        # Update m2m through field
        self._update_nested(instance, "custom_styles", CustomStyleSerializer)
        self._update_m2m_through(instance, "fields", FilterFieldSerializer)

        return instance

    def to_internal_value(self, data):
        data["group"], data["name"] = self._get_layer_group(data)
        return super().to_internal_value(data)

    def to_representation(self, obj):
        return {
            **super().to_representation(obj),
            "name": self._get_name_path(obj),
            "view": obj.group.view,
        }

    def _get_layer_group(self, data):
        view = data["view"]

        try:
            group_path, layer_name = data["name"].rsplit("/", 1)
        except ValueError:
            group_path, layer_name = "Unknown", data["name"]

        group = None
        for group_name in group_path.split("/"):
            if group:
                group, _ = group.children.get_or_create(label=group_name)
            else:
                group, _ = LayerGroup.objects.get_or_create(view=view, label=group_name)

        return group.pk, layer_name

    def _get_name_path(self, obj):
        def get_group_path(group):
            name = group.label
            if group.parent:
                name = get_group_path(group.parent) + f"/{name}"
            return name

        group_path = get_group_path(obj.group)
        return f"{group_path}/{obj.name}"

    @transaction.atomic
    def update(self, instance, validated_data):

        instance = super().update(instance, validated_data)

        # Update m2m through field
        self._update_m2m_through(instance, "fields", FilterFieldSerializer)
        self._update_nested(instance, "custom_styles", CustomStyleSerializer)

        return instance

    def _update_nested(self, instance, field, serializer):
        getattr(instance, field).all().delete()

        for value in self.initial_data.get(field, []):
            obj = serializer(data=value)
            if obj.is_valid(raise_exception=True):
                obj.save(layer=instance)

    def _update_m2m_through(self, instance, field, serializer):
        getattr(instance, field).clear()

        for value in self.initial_data.get(field, []):
            try:
                value["field"] = value["id"]
            except KeyError:
                raise ValidationError("Fields must contain Source's field id")

            obj = serializer(data=value)
            if obj.is_valid(raise_exception=True):
                obj.save(layer=instance)

    class Meta:
        model = Layer
        fields = "__all__"



from django.contrib.auth import get_user_model
from django.test import TestCase
from django.urls import reverse
from rest_framework.status import HTTP_200_OK, HTTP_201_CREATED
from rest_framework.test import APIClient

from django_geosource.models import PostGISSource, FieldTypes
from terra_layer.models import Layer, LayerGroup, FilterField

UserModel = get_user_model()


class ModelSourceViewsetTestCase(TestCase):
    def setUp(self):
        self.client = APIClient()
        self.default_user = UserModel.objects.get_or_create(
            is_superuser=True, **{UserModel.USERNAME_FIELD: "testuser"}
        )[0]
        self.client.force_authenticate(self.default_user)

        self.source = PostGISSource.objects.create(
            name="test_view",
            db_name="test",
            db_password="test",
            db_host="localhost",
            geom_type=1,
            refresh=-1,
        )

    def test_list_view(self):
        # Create many sources and list them
        group = LayerGroup.objects.create(view=0, label="Test Group")

        [Layer.objects.create(group=group, source=self.source) for x in range(5)]

        response = self.client.get(reverse("terralayer:layer-list"))
        self.assertEqual(response.status_code, HTTP_200_OK)
        self.assertEqual(Layer.objects.count(), len(response.json()))

    def test_create_layer(self):
        query = {
            "source": self.source.pk,
            "view": 0,
            "name": "test layer",
            "table_export_enable": True,
            "filter_enable": False,
        }

        response = self.client.post(reverse("terralayer:layer-list"), query)
        self.assertEqual(HTTP_201_CREATED, response.status_code)

        response = response.json()

        self.assertTrue(response.get("table_export_enable"))
        self.assertFalse(response.get("filter_enable"))

    def test_update_layer(self):
        group = LayerGroup.objects.create(view=0, label="Test Group")

        field = self.source.fields.create(
            name="test_field", label="test_label", data_type=FieldTypes.String.value
        )
        layer = Layer.objects.create(
            group=group, source=self.source, minisheet_enable=False
        )
        FilterField.objects.create(
            label="test layer fields",
            layer=layer,
            field=field,
            filter_settings={},
            filter_enable=True,
        )

        query = {
            "source": self.source.pk,
            "view": 10,
            "name": "test layer",
            "minisheet_enable": True,
            "filter_enable": True,
        }

        response = self.client.patch(
            reverse("terralayer:layer-detail", args=[layer.pk]), query
        )

        self.assertEqual(response.status_code, HTTP_200_OK)

        response = response.json()
        self.assertTrue(response.get("minisheet_enable"))
